#include <arpa/inet.h>
#include <errno.h>
#include <pthread.h>
#include <stdio.h>
#include <string.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <unistd.h>
#include "net.h"


/*
*    Lookup telephone number AGID.
*    Required external program is curl
*
*    Author: R.J. van der Putten, Leiden, Holland, 
*    rob at sput dot nl.
* 
*    This program is free software; you can redistribute it and/or modify
*    it under the terms of the GNU General Public License version 3 as
*    published by the Free Software Foundation.
* 
*    This program is distributed in the hope that it will be useful,
*    but WITHOUT ANY WARRANTY; without even the implied warranty of
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
*    GNU General Public License for more details.
*/


#pragma GCC diagnostic ignored "-Wpragmas"
#pragma GCC diagnostic ignored "-Wstringop-truncation"
#pragma GCC diagnostic ignored "-Wformat-truncation"


/*
*  Function list
*
*  Open listening socket; openlsock()
*  Proces listening socket; proclsock()
*/


/*  Open listening socket  */
int openlsock(struct sockaddr_in6 *addr, uint16_t port, int backlog)
{
    char portstr[8];
    int opt6only, optreuse, sfd;
    socklen_t addrlen;

    memset(portstr, 0, 8);

    opt6only =  0;
    optreuse =  1;
    sfd      = -1;

    snprintf(portstr, 7, "%d", port);
    wrtsysl("openlsock: Opening listening socket port", portstr, 0);

    /*   Open IPv6 TCP socket  */
    if ((sfd = socket(AF_INET6, SOCK_STREAM | SOCK_NONBLOCK, 0)) < 0) {
	wrtsysl("openlsock(): Error socket", strerror(errno), 1);
	return(-1);
    }
    /*  Set option to reuse  */
    if ((setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &optreuse, \
	  sizeof(optreuse))) < 0) {
	wrtsysl("openlsock(): Error setsockopt so_reuse=1", strerror(errno), 1);
	return(-1);
    }
    /*  Set to IPv6 before Dual stack option  */
    memset(addr, 0, sizeof(struct sockaddr_in6));
    addr->sin6_family = AF_INET6;
    /*  Now set option to dual stack  */
    if ((setsockopt(sfd, IPPROTO_IPV6, IPV6_V6ONLY, &opt6only, \
      sizeof(opt6only))) < 0) {
	wrtsysl("openlsock(): Error setsockopt ipv6_v6only=0", \
	  strerror(errno), 1);
	return(-1);
    }
    /*  Bind  */
    addr->sin6_port = htons(port);
    inet_pton(AF_INET6, "::", &(addr->sin6_addr));
    addrlen = sizeof(struct sockaddr_in6);
    if ((bind(sfd, (struct sockaddr *) addr, addrlen)) < 0) {
	wrtsysl("openlsock(): Error bind", strerror(errno), 1);
	return(-1);
    }
    /*  Listen  */
    if (listen(sfd, backlog) < 0) {
	wrtsysl("openlsock(): Error listen", strerror(errno), 1);
	return(-1);
    }

    return(sfd);
}

/*  Proces listening socket  */
void * proclsock(void *idnum)
{
    int i, id, *idptr, inp, len, *lktyp, outp, ret, todo;
    volatile int *cont;
    char *buf, *lnam, *line, *logstr, *name, *num;
    wchar_t *wnam;

    i    = 0;
    inp  = 0;	/*  Current position in input buffer  */
    len  = 0;
    outp = 0;	/*  Current position in output buffer  */
    ret  = 0;	/*  Bytes in buf  */
    todo = 0;	/*  Bytes yet to write  */

    /*  My own thread number (not a thrdid)  */
    idptr = (int *) idnum;
    id = *idptr;

    /*  Continue while not zero  */
    cont = &(thrdtab[id].cont);

    /*  Lookup type  */
    lktyp = &(thrdtab[id].lktyp);

    /*  Buffers  */
    buf  = thrdtab[id].bufs[0];
    line = thrdtab[id].bufs[1];
    lnam = thrdtab[id].lnam;
    logstr = thrdtab[id].logstr;
    name = thrdtab[id].phone.name;
    num  = thrdtab[id].phone.num;
    wnam = thrdtab[id].wnam;
    buf[0]  = 0;
    line[0] = 0;
    lnam[0] = 0;
    name[0] = 0;
    num[0]  = 0;
    wnam[0] = 0;

    if (debflg != 0) {
	snprintf(logstr, 255, "Debug: proclsock(): thread nr: %d id: %lx", \
	  id, thrdtab[id].thrdid);
	wrtolog(logstr, NULL);
    }

    /*  Read from socket  */
    outp = 0;
    while (*cont != 0 && line[0] != 10 && line[0] != 13) {
	ret = read(thrdtab[id].sfd, buf, ANLD_BUF_MAX);
	if (ret < 0) {
	    if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
		if (*cont == 0) {
		    wrtsysl("proclsock: Thread stopped", NULL, 0);
		    break;
		} else {
		    ret = 0;
		    continue;
		}
	    } else {
		wrtsysl("proclsock(): Cant read", NULL, 1);
		*cont = 0;
		break;
	    }
	} else if (ret == 0) {
	    break;
	}
	/*  Find lines of text  */
	inp = 0;
	buf[ret] = 0;
	while (inp < ret) {
	    line[outp] = buf[inp];
	    if (line[outp] == 10) {
		/*  A line  */
		outp++;
		line[outp] = 0;
		if (debflg != 0) {
		    fprintf(flog, "%s", line);
		    fflush(flog);
		}
		/*  Evaluate contents of line  */
		if (strncasecmp(line, "agi_callerid: ", 14) == 0) {
		    strncpy(num, line + 14, ANLD_NUM_MAX);
		    num[ANLD_NUM_MAX] = 0;
		} else if (strncasecmp(line, "agi_calleridname: ", 18) == 0) {
		    strncpy(name, line + 18, ANLD_NAME_MAX);
		    name[ANLD_NAME_MAX] = 0;
		} else if (line[0] == 10 || line[0] == 13) {
		    /*  Empty line: Done  */
		    break;
		}
		outp = 0;
		line[outp] = 0;
	    } else {
		outp++;
		line[outp] = 0;
	    }
	    if (outp > ANLD_BUF_MAX2) {
		wrtsysl("proclsock(): Line too long", NULL, 1);
		*cont = 0;
		break;
	    }
	    inp++;
	}
    }

    if (*cont == 0) {
	/*  Premature end  */
	if (thrdtab[id].weblkp.pweb != NULL) {
	    pclose(thrdtab[id].weblkp.pweb);
	    thrdtab[id].weblkp.pweb = NULL;
	}
	close(thrdtab[id].sfd);
	wrtsysl("proclsock: Thread stopped", NULL, 0);
	thrdtab[id].sfd = -1;
	return(NULL);
    }

    /*  Cleanup number  */
    i = 0;
    while (i < ANLD_NUM_SIZ) {
	if (num[i] < ' ' && num[i] > -1) {
	    num[i] = 0;
	    break;
	}
	i++;
    }
    len = i;	/*  Length of number  */

    /*  Cleanup name  */
    i = 0;
    while (i < ANLD_NAME_SIZ) {
	if (name[i] < ' ' && name[i] > -1) {
	    name[i] = 0;
	    break;
	}
	i++;
    }

    if (debflg != 0) {
	if(num[0] != 0)
	    wrtolog("Debug: proclsock(): found num", num);
	if(name[0] != 0)
	    wrtolog("Debug: proclsock(): found name", name);
    }

    /*  Blacklist check  */
    if (name[0] != 0) { 
	/*  Convert name to lower case and lookup  */
	if (blkdb != NULL && blkent != 0 && namtolow(name, lnam, wnam) == 0) {
	    /*  Blocklist present and namtolow() OK  */
	    if (blklookup(lnam) != 0) {
		/*  lnam found; clear name  */
		name[0] = 0;
	    }
	} else {
	    /*  namtolow() failed or no name blocklist file  */
	    if (strcasecmp(name, "unknown") == 0 || \
	      strcasecmp(name, "Privacy Manager") == 0 || \
	      strcasecmp(name, "Anonymous") == 0) {
		/*  Clear  */
		wrtsysl("proclsock: Invalid", name, 0);
		name[0] = 0;
	    }
	}
    }
    if(name[0] != 0)
	wrtsysl("proclsock: Name from ITSP", name, 0);

    /*  Check for number and no name  */
    if (num[0] == 0 || name[0] != 0) {
	/*  Nothing to do  */
	if (thrdtab[id].weblkp.pweb != NULL) {
	    pclose(thrdtab[id].weblkp.pweb);
	    thrdtab[id].weblkp.pweb = NULL;
	}
	close(thrdtab[id].sfd);
	thrdtab[id].sfd = -1;
	*cont = 0;
	return(NULL);
    }

    /*  Check for number; 0123456789  */
    i = 0;
    while (i < len) {
	if (num[i] < '0' || num[i] > '9') {
	    /*  Invalid number  */
	    if (thrdtab[id].weblkp.pweb != NULL) {
		pclose(thrdtab[id].weblkp.pweb);
		thrdtab[id].weblkp.pweb = NULL;
	    }
	    close(thrdtab[id].sfd);
	    thrdtab[id].sfd = -1;
	    *cont = 0;
	    return(NULL);
	}
	i++;
    }

    /*  Write number to syslog  */
    wrtsysl("proclsock: Lookup nr", num, 0);

    if (numbers.idx != NULL && numbers.ent != 0) {
	/*  Numbers database exists; Lookup name in database  */
	*lktyp = 0;
	phndblookup(num, name, numbers);
    }
    if (name[0] == 0 && disweb == 0) {
	/*  Not found; Do a web lookup  */
	*lktyp = 1;
	weblookup(id);
	/*  2nd blacklist check  */
	if (name[0] != 0) {
	    /*  Convert name to lower case and lookup  */
	    if (blkdb != NULL && blkent != 0 && \
	      namtolow(name, lnam, wnam) == 0) {
		/*  Blocklist present and namtolow() OK  */
		if (blklookup(lnam) != 0) {
		    /*  lnam found; clear name  */
		    name[0] = 0;
		}
	    } else {
		/*  namtolow() failed or no name blocklist file  */
		if (strcasecmp(name, "unknown") == 0 || \
		  strcasecmp(name, "Privacy Manager") == 0 || \
		  strcasecmp(name, "Not Found") == 0 || \
		  strcasecmp(name, "Anonymous") == 0) {
		    /*  Clear  */
		    wrtsysl("proclsock: Invalid", name, 0);
		    name[0] = 0;
		}
	    }
	}
    }	/*  End of if name[0] == 0 && disweb == 0  */

    if (name[0] == 0 && areas.idx != NULL && areas.ent != 0) {
	/*  Not found and areas database exists; Do a area code lookup  */
	*lktyp = 2;
	phndblookup(num, name, areas);
    }
    if (name[0] == 0 && regions.idx != NULL && regions.ent != 0) {
	/*  Not found and regions database exists; Do a region code lookup  */
	*lktyp = 3;
 	phndblookup(num, name, regions);
    }

    if (name[0] != 0) {
	/*  Found name or location  */
	if (*lktyp > 1) {
	    /*  Geographic  */
	    snprintf(buf, ANLD_BUF_MAX2, "SET VARIABLE NAME \"Lc %s\"\n", name);
	} else {
	    /*  Name  */
	    snprintf(buf, ANLD_BUF_MAX2, "SET VARIABLE NAME \"%s\"\n", name);
	}
	/*  Write to socket  */
	len = strlen(buf);	/*  Length of string to write  */
	todo = len;
	while (*cont != 0 && todo > 0) {
	    ret = write(thrdtab[id].sfd, buf + len - todo, todo);
	    if (ret < 0) {
		if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
		    if (*cont == 0) {
			wrtsysl("proclsock: Thread stopped", NULL, 0);
			break;
		    } else {
			ret = 0;
		    }
		} else {
		    wrtsysl("proclsock(): Cant write", NULL, 1);
		    *cont = 0;
		    break;
		}
	    }
	    todo = todo - ret;
	}

	/*  Write to log  */
	snprintf(logstr, 255, "%s\t%s\t%s\t", lookups[*lktyp], num, name);
	wrtolog(logstr, NULL);

	if (*cont == 0) {
	    /*  Premature end  */
	    if (thrdtab[id].weblkp.pweb != NULL) {
		pclose(thrdtab[id].weblkp.pweb);
		thrdtab[id].weblkp.pweb = NULL;
	    }
	    close(thrdtab[id].sfd);
	    wrtsysl("proclsock: Thread stopped", NULL, 0);
	    thrdtab[id].sfd = -1;
	    return(NULL);
	}

	/*  Read from socket  */
	outp = 0;
	while (*cont != 0) {
	    ret = read(thrdtab[id].sfd, buf, ANLD_BUF_MAX);
	    if (ret < 0) {
		if (errno == EINTR || errno == EAGAIN || errno == EWOULDBLOCK) {
		    if (*cont == 0) {
			wrtsysl("proclsock: Thread stopped", NULL, 0);
			break;
		    } else {
			ret = 0;
			continue;
		    }
		} else {
		    wrtsysl("proclsock(): Cant read", NULL, 1);
		    *cont = 0;
		    break;
		}
            } else if (ret == 0) {
        	break;
	    }
            
	    /*  Find line  */
	    inp = 0;
	    buf[ret] = 0;
	    while (inp < ret) {
		line[outp] = buf[inp];
		if (line[outp] == 10) {
		    /*  A line  */
		    outp++;
		    line[outp] = 0;
		    /*  Done  */
		    *cont = 0;
		    break;
		} else {
		    outp++;
		    line[outp] = 0;
		}
		if (outp > ANLD_BUF_MAX2) {
		    wrtsysl("proclsock(): Line too long", NULL, 1);
		    *cont = 0;
		    break;
		}
		inp++;
	    }
	}
    }

    if (*cont != 0) {
	shutdown(thrdtab[id].sfd, SHUT_RDWR);
	/*
	*  Wait before close().
	*  If this time is too short, increase in common.h
	*/
	usleep(ANLD_CLS_SOCK_TIM);
    }
    close(thrdtab[id].sfd);
    thrdtab[id].sfd = -1;
    *cont = 0;

    if (debflg != 0) {
	snprintf(logstr, 255, "Debug: proclsock(): thread nr: %d ended", id);
	wrtolog(logstr, NULL);
    }

    return(NULL);
}

